#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import os
import re
import sys

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time, TimeDelta

from . import archiveutils as au

"""
Represents a VLBA IDIFITS file
"""

LOG = logging.getLogger("ingestion")


class IDIFits:
    def __init__(self, file):
        self.idifitsfile = file
        self.hdu = fits.open(file)
        self.source_table = Table(self.hdu["SOURCE"].data)
        self.uv_data = self.hdu["UV_DATA"].data
        self.frequency_data = self.hdu["FREQUENCY"].data
        self.frequency_table = Table(self.hdu["FREQUENCY"].data)
        self.reference_frequency = self.hdu["FREQUENCY"].header["REF_FREQ"]
        self.numbands = self.hdu["FREQUENCY"].header["NO_BAND"]
        # self.spectchannels = self.hdu['FREQUENCY'].header['NO_CHAN']
        self.scans = self.build_scans(self.uv_data)
        self.hdu.close()

    def getfitsfile(self):
        return self.idifitsfile

    def getproject(self):
        m = re.match("([A-Z]+\d+)", self.hdu["PRIMARY"].header["OBSERVER"])
        return m.group(0)

    def getexperiment(self):
        m = re.match("[A-Z]+\d+([A-Z|\d]+)", self.hdu["PRIMARY"].header["OBSERVER"])
        if m is None:
            return ""
        return m.group(1)

    def getcorrelatorpass(self):
        fflist = os.path.basename(self.getfitsfile()).split("_")
        return fflist[2]

    def getsourcetable(self):
        return self.source_table

    def getsource(self, source):
        """
        Return source name, RA, and DEC of scan source
        :param source: the source index of the scan source
        :return: source name, ra, dec
        """
        sourcekey = self.hdu["SOURCE"].data.columns[0].name
        source_row = self.source_table[self.source_table[sourcekey] == source]
        # Retrieving a column with 1 element so extract the value
        # return source_row[0]['SOURCE'], source_row[0]['RAOBS'], source_row[0]['DECOBS']
        return source_row[0]["SOURCE"], source_row[0]["RAAPP"], source_row[0]["DECAPP"]

    def getfrequency(self, frequency):
        frequency_row = self.frequency_table[self.frequency_table["FREQID"] == frequency]
        frequencies = frequency_row[0]["BANDFREQ"]
        if np.isscalar(frequencies):
            # If the frequency is a scalar, put it in a numpy ndarray
            freq_array = np.ndarray(shape=(1))
            freq_array[0] = frequency_row[0]["BANDFREQ"]
            return freq_array
        return frequencies

    def getbandwidth(self, frequency):
        frequency_row = self.frequency_table[self.frequency_table["FREQID"] == frequency]
        if np.isscalar(frequency_row[0]["TOTAL_BANDWIDTH"]):
            return frequency_row[0]["TOTAL_BANDWIDTH"]
        else:
            return frequency_row[0]["TOTAL_BANDWIDTH"][0]

    def getfrequencytable(self):
        return self.frequency_table

    def get_number_bands(self):
        return self.numbands

    # def get_num_spectral_channels(self):
    #     return self.spectchannels

    def build_scans(self, uv_data):
        if len(uv_data) <= 0:
            LOG.critical(f"UV_DATA table is empty for VLBA file: {self.idifitsfile}")
            sys.exit(1)
        sourcekey = self.hdu["SOURCE"].data.columns[0].name
        scan_list = []
        polarizations = self.getpolarizations(self.hdu["UV_DATA"].header)

        source = uv_data[0]["SOURCE"]
        scan = 0
        scan_start_row = 0
        frequency_ids = [uv_data[scan_start_row]["FREQID"]]
        for row in range(len(uv_data)):
            if source != uv_data[row]["SOURCE"]:
                scan_end_row = row - 1
                scan_start_time = Time(uv_data[scan_start_row]["DATE"], format="jd") + TimeDelta(
                    uv_data[scan_start_row]["TIME"], format="jd"
                )
                scan_end_time = Time(uv_data[scan_end_row]["DATE"], format="jd") + TimeDelta(
                    uv_data[scan_end_row]["TIME"], format="jd"
                )
                scan_time_on_source = scan_end_time - scan_start_time
                scan_integration_time = uv_data[scan_start_row]["INTTIM"]
                scan_source = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]][
                    "SOURCE"
                ][0]
                source_ra = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]][
                    "RAAPP"
                ][0]
                source_dec = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]][
                    "DECAPP"
                ][0]
                # c = SkyCoord(source_ra, source_dec, frame='icrs', unit='deg')
                # source_radec = c.to_string('hmsdms')
                source_radec = ""
                frequencies = [(self.get_reference_frequency() + self.getfrequency(f)) for f in frequency_ids]
                frequency_list = np.concatenate(frequencies, axis=0) / 1000000
                total_bandwidth = self.getbandwidth(1)
                current_scan = [
                    scan,
                    scan_source,
                    scan_start_time.mjd,
                    scan_end_time.mjd,
                    scan_time_on_source.sec,
                    scan_integration_time,
                    " ".join(map(str, frequency_list)),
                    total_bandwidth,
                    source_ra,
                    source_dec,
                    source_radec,
                    polarizations,
                ]
                scan_list.append(current_scan)
                scan = scan + 1
                scan_start_row = row
                frequency_ids.clear()

            frequency_id = uv_data[row]["FREQID"]
            if frequency_id not in frequency_ids:
                frequency_ids.append(frequency_id)

            source = uv_data[row]["SOURCE"]

        # Add in final scan
        scan_end_row = row
        scan_start_time = Time(uv_data[scan_start_row]["DATE"], format="jd") + TimeDelta(
            uv_data[scan_start_row]["TIME"], format="jd"
        )
        scan_end_time = Time(uv_data[scan_end_row]["DATE"], format="jd") + TimeDelta(
            uv_data[scan_end_row]["TIME"], format="jd"
        )
        scan_time_on_source = scan_end_time - scan_start_time
        scan_integration_time = uv_data[scan_start_row]["INTTIM"]
        scan_source = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]]["SOURCE"][0]
        source_ra = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]]["RAAPP"][0]
        source_dec = self.source_table[self.source_table[sourcekey] == uv_data[scan_start_row]["SOURCE"]]["DECAPP"][0]
        # c = SkyCoord(source_ra, source_dec, frame='icrs', unit='deg')
        # source_radec = c.to_string('hmsdms')
        source_radec = ""
        frequencies = [(self.get_reference_frequency() + self.getfrequency(f)) for f in frequency_ids]
        frequency_list = np.concatenate(frequencies, axis=0) / 1000000
        total_bandwidth = self.getbandwidth(1)
        current_scan = [
            scan,
            scan_source,
            scan_start_time.mjd,
            scan_end_time.mjd,
            scan_time_on_source.sec,
            scan_integration_time,
            " ".join(map(str, frequency_list)),
            total_bandwidth,
            source_ra,
            source_dec,
            source_radec,
            polarizations,
        ]
        scan_list.append(current_scan)

        # Build scan table from list of scans
        return Table(
            rows=scan_list,
            names=(
                "SCAN_ID",
                "SOURCE",
                "START_TIME",
                "END_TIME",
                "TIME_ON_SOURCE",
                "INTEGRATION_TIME",
                "FREQUENCY",
                "BANDWIDTH",
                "RA",
                "DEC",
                "RA/DEC",
                "POLARIZATION",
            ),
        )

    def get_reference_frequency(self):
        return self.reference_frequency

    def getpolarizations(self, uv_header):
        numpolarizations = uv_header["NO_STKD"]
        polarization_code = 0
        for i in range(numpolarizations):
            pol_key = f"STK_{i+1}"
            if pol_key not in uv_header.keys():
                LOG.warning(f"Stokes key: {pol_key} is not present in UV_DATA table")
                continue
            polarization = self.lookup_polarization(uv_header[pol_key])
            polarization_code = polarization_code + polarization
        return au.getpolarization(polarization_code)

    def lookup_polarization(self, polarization_code):
        # Map of polarization codes used by FITS to observation polarization codes
        polarization_codes = {
            1: "I",
            2: "Q",
            3: "U",
            4: "V",
            -1: "RR",
            -2: "LL",
            -3: "RL",
            -4: "LR",
            -5: "XX",
            -6: "YY",
            -7: "XY",
            -8: "YX",
        }

        if polarization_code not in polarization_codes.keys():
            LOG.error(f"Invalid polarization code: {polarization_code}")
            sys.exit(1)

        return au.getpolarizationcode(polarization_codes[polarization_code])

    def getuvdatacolumns(self):
        return self.uv_data.columns

    def gethdulist(self):
        return self.hdu.info()

    def getheader(self):
        return self.hdu["PRIMARY"].header

    def getobservingband(self):
        ref_frequency = self.get_reference_frequency()
        return self.lookup_observing_band(ref_frequency)

    def getfrequencyoffset(self, freq_id):
        frequency_row = self.frequency_table[self.frequency_table["FREQID"] == freq_id]
        return frequency_row["BANDFREQ"][0][1]

    def getbandwidthoffset(self, freq_id):
        frequency_row = self.frequency_table[self.frequency_table["FREQID"] == freq_id]
        return frequency_row["CH_WIDTH"][0][1]

    def lookup_observing_band(self, reference_frequency):
        ref_freq = reference_frequency / 1000000000
        obs_band = ""

        if 0.058 <= ref_freq <= 0.084:
            obs_band = "4"
        elif 0.236 <= ref_freq <= 0.492:
            obs_band = "P"
        elif 1.0 <= ref_freq <= 2.0:
            obs_band = "L"
        elif 2.0 < ref_freq <= 4.0:
            obs_band = "S"
        elif 4.0 < ref_freq <= 8.0:
            obs_band = "C"
        elif 8.0 < ref_freq <= 12.0:
            obs_band = "X"
        elif 12.0 < ref_freq <= 18.0:
            obs_band = "Ku"
        elif 18.0 < ref_freq <= 26.5:
            obs_band = "K"
        elif 26.5 < ref_freq <= 40.0:
            obs_band = "Ka"
        elif 40.0 < ref_freq <= 50.0:
            obs_band = "Q"
        else:
            LOG.warning(f"Reference frequency {ref_freq} could not be located in a band")

        return obs_band

    def dump_scans(self):
        self.scans.pprint(max_lines=-1, max_width=-1)


def test_1(file):
    pass


if __name__ == "__main__":
    print("Main")
    # test_1(os.path.join(vlba_stage, 'VLBA_BS263E_bs263e_BIN0_SRC0_0_180716T144209.idifits'))
